// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "../Asset.hpp"
#include "../AssetLoader.hpp"
#include "display/font/TrueTypeFont.hpp"

namespace Dawn {
  class TrueTypeAsset : public Asset {
    protected:
      AssetLoader loader;
      uint8_t *buffer = nullptr;
      truetypechar_t *characterData = nullptr;
      uint8_t *pixels = nullptr;
      int32_t width, height;

    public:
      TrueTypeFont font;

      /**
       * Constructs a new True Type Asset. As with all other assets you should
       * instead use the AssetManaager.load method.
       * 
       * @param assMan Asset manager that this asset belongs to.
       * @param name Filename of this asset.
       */
      TrueTypeAsset(AssetManager *assMan, std::string name);

      void updateSync() override;
      void updateAsync() override;

      /**
       * Disposes / Cleans up the truetype asset.
       */
      ~TrueTypeAsset();
  };
}