// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "display/font/truetype/NewTrueTypeShared.hpp"
#include "util/mathutils.hpp"
#include "display/Texture.hpp"
#include "util/UsageLock.hpp"

namespace Dawn {
  class NewTrueTypeAsset;

  struct NewTrueTypeCharacter {
    float_t advanceX;
    float_t advanceY;
    glm::vec2 bitmapSize;
    glm::vec2 bitmapPosition;
    float_t textureY;
  };

  struct NewTrueTypeFaceTextureStyle {
    uint32_t fontSize;
    flag_t style;

    bool operator < (const struct NewTrueTypeFaceTextureStyle& r) const {
      return std::tie(fontSize, style) < std::tie(r.fontSize, r.style);
    }
  };

  class NewTrueTypeFaceTexture {
    public:
      FT_Face face;
      std::map<FT_ULong, struct NewTrueTypeCharacter> characterData;
      struct NewTrueTypeFaceTextureStyle style;
      UsageLock locks;
      Texture texture;

      /**
       * Construct a new New True Type Face Texture object
       * 
       * @param face The freetype face object.
       * @param style Style that this font has, used for locking.
       */
      NewTrueTypeFaceTexture(
        FT_Face face,
        struct NewTrueTypeFaceTextureStyle style
      );

      /**
       * Returns the character data for the given character.
       * 
       * @param c Character to get data for.
       * @return The Character data for the given character.
       */
      struct NewTrueTypeCharacter getCharacterData(FT_ULong c);

      /**
       * Destroys this true type face texture.
       */
      ~NewTrueTypeFaceTexture();

      friend class NewTrueTypeAsset;
  };
}