// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TrueTypeTool.hpp"
#ifndef  STB_TRUETYPE_IMPLEMENTATION
  #define STB_TRUETYPE_IMPLEMENTATION
  #include <stb_truetype.h>
#endif

using namespace Dawn;

std::vector<std::string> TrueTypeTool::getRequiredFlags() {
  return { "input", "output" };
}

int32_t TrueTypeTool::start() {
  File fileIn(flags["input"]);
  char *ttfData = nullptr;
  if(!fileIn.readToBuffer(&ttfData)) {
    std::cout << "Failed to read file to buffer!" << std::endl;
    return 1;
  }

  // Create bitmap data. This is a single channel color (alpha).
  size_t width = 1024;
  size_t height = 1024;
  float_t fontSize = 128.0f;
  size_t textureSize = width * height;
  stbi_uc *bitmapData = new stbi_uc[textureSize];
  stbtt_bakedchar characterData[TRUETYPE_NUM_CHARS];

  // Now parse the TTF itself.
  stbtt_BakeFontBitmap(
    (uint8_t*)ttfData, 0, (float)fontSize, bitmapData,
    width, height,
    TRUETYPE_FIRST_CHAR, TRUETYPE_NUM_CHARS,
    characterData
  );

  // Prepare output file for writing.
  File fileOut(flags["output"] + ".truetype");
  if(!fileOut.mkdirp()) {
    free(ttfData);
    delete bitmapData;
    std::cout << "Failed to create output directory!" << std::endl;
    return 1;
  }
  if(!fileOut.open(FILE_MODE_WRITE)) {
    free(ttfData);
    delete bitmapData;
    std::cout << "Failed to open output file for writing!" << std::endl;
    return 1;
  }

  // Write data
  fileOut.writeString(std::to_string(width) + "|" + std::to_string(height) + "|" + std::to_string(fontSize) + "|");
  uint8_t pixels[1];
  for(size_t i = 0; i < textureSize; i++) {
    pixels[0] = bitmapData[i];
    fileOut.writeRaw((char*)pixels, 1);
  }

  // Write quads
  fileOut.writeRaw((char*)characterData, sizeof(stbtt_bakedchar) * TRUETYPE_NUM_CHARS);
  free(ttfData);
  delete bitmapData;
  fileOut.close();
  
  return 0;
}